// sched.h
#ifndef SCHED_H
#define SCHED_H

#include "string.h"
#include "utils.h"
#include "types.h"
#include "memory.h"

#define MAX_PROCESSES 256
#define MAX_THREADS 16
#define QUEUE_SIZE 256
#define TIME_SLICE 10
#define PROCESS_PAGE_TABLE_INDEX 0
#define THREAD_STACK_SIZE 0x1000 // 使用一个完整的页(4KB)作为线程栈
#define MAX_PRIORITY 10
#define MIN_PRIORITY 1
#define AGING_THRESHOLD 1000 // 老化阈值
#define PRIORITY_BOOST 1     // 每次老化提升的优先级

enum task_state
{
    TASK_RUNNING,
    TASK_READY,
    TASK_BLOCKED,
    TASK_SLEEPING
};

// 线程栈结构
struct thread_stack
{
    void *stack_page;   // 指向分配的页面
    uint32_t stack_top; // 栈顶位置
};

// 线程控制块
struct thread
{
    unsigned int esp;
    unsigned int ebp;
    unsigned int eip;
    int time_slice;
    enum task_state state;
    int thread_id;
    struct thread_stack stack; // 线程的栈信息
    struct task *parent_process;
    int priority;
    int initial_priority; // 初始优先级
    int wait_time;        // 等待时间
};

// 进程控制块
struct task
{
    int pid;
    int priority;
    int remaining_time;
    uint32_t page_table_phys_addr;
    struct thread threads[MAX_THREADS];
    int thread_count;
    void *process_memory; // 进程的共享内存页面
};

typedef struct
{
    int front, rear;
    struct thread *threads[QUEUE_SIZE];
} ThreadQueue;

typedef struct semaphore
{
    int value;
    ThreadQueue *waiters;
    int highest_priority_waiter;
} semaphore_t;

// 队列操作函数声明
void init_thread_queue(void);
int enqueue_thread(struct thread *thread);
struct thread *dequeue_thread(void);

// 线程管理函数
int create_thread(int pid, unsigned int eip, int priority, void *arg);
void terminate_thread(int pid, int thread_id);
void yield_thread(void);

// 进程管理函数
int add_task(unsigned int eip, int priority, int remaining_time);
void terminate_task(int pid);

// 调度相关函数
void schedule(void);
void timer_tick(void);
void block_current_thread(void);
void unblock_thread(struct thread *thread);
void print_task_status(void);
void update_thread_priority(void);

// 同步原语
void semaphore_init(semaphore_t *sem, int value);
void semaphore_wait(semaphore_t *sem);
void semaphore_signal(semaphore_t *sem);

// 外部函数声明
extern unsigned int get_esp(void);
extern unsigned int get_ebp(void);
extern unsigned int get_eip(void);
extern void switch_context(uint32_t new_esp, uint32_t new_ebp, uint32_t new_eip);

extern ThreadQueue thread_queue;

#endif