#include "memory.h"
#include "types.h"
#include "utils.h"
#include "sched.h"

uint64_t *free_page_bitmap;

uint64_t total_pages;

uint64_t *page_access_time;

semaphore_t mem_alloc_sem;

// Initialize the memory system
void memory_init(void)
{
    setup_paging();
    total_pages = 1024; //4MB = 1024 * 4096

    // 为位示图分配一个页
    free_page_bitmap = (uint64_t *)0x4000;                                
    page_access_time = (uint64_t *)(free_page_bitmap + total_pages / 64); // 追踪每个页面上次使用时间，用于LRU

    // 初始化位示图（全空）
    for (uint64_t i = 0; i < total_pages / 64; i++)
    {
        free_page_bitmap[i] = 0xFFFFFFFFFFFFFFFF;
    }

    // 保留前几页（内核等）
    uint64_t reserved_pages = 16;
    for (uint64_t i = 0; i < reserved_pages; i++)
    {
        uint64_t bitmap_index = i / 64;
        uint64_t bit_offset = i % 64;
        free_page_bitmap[bitmap_index] &= ~(1ULL << bit_offset);
    }

    // 初始化页面访问时间
    for (uint64_t i = 0; i < total_pages; i++)
    {
        page_access_time[i] = 0;
    }
    semaphore_init(&mem_alloc_sem, 1); // 信号量初始化（临界区：页面操作）
}

void *alloc_page(void)
{
    semaphore_wait(&mem_alloc_sem); // 请求内存分配信号量

    while (1)
    {
        for (uint64_t i = 0; i < total_pages / 64; i++)
        {
            if (free_page_bitmap[i] != 0)
            {
                int page_bit = ffs_custom(free_page_bitmap[i]) - 1;
                if (page_bit != -1)
                {
                    free_page_bitmap[i] &= ~(1ULL << page_bit);
                    update_page_access_time(page_bit);
                    uintptr_t page_addr = (uintptr_t)(page_bit * PAGE_SIZE);
                    semaphore_signal(&mem_alloc_sem); // 释放内存分配信号量
                    return (void *)page_addr;
                }
            }
        }
        // 如果没有找到可用页面，则尝试通过LRU算法驱逐页面
        lru_evict_page();
    }
}

//分配页面并清空该页面
void *os_alloc_page(void)
{
    void *page = alloc_page();
    if (page)
    {
        memset(page, 0, 0x1000); // Clear the page
    }
    return page;
}

// 释放页面
void free_page(void *ptr)
{
    semaphore_wait(&mem_alloc_sem);
    uint64_t page_addr = (uint64_t)((uintptr_t)ptr);
    uint64_t page_index = page_addr / 0x1000;
    uint64_t bitmap_index = page_index / 64;
    uint64_t bit_offset = page_index % 64;

    free_page_bitmap[bitmap_index] |= (1ULL << bit_offset);

    page_access_time[page_index] = 0;
    semaphore_signal(&mem_alloc_sem);
}

// 更新页面访问时间
void update_page_access_time(uint64_t page_index)
{
    if (!(page_access_time[page_index] & PAGE_LOCKED))
    {
        static uint64_t current_time = 0;
        if (++current_time == MAX_TIME)
        {
            current_time = 0;
            for (uint64_t i = 0; i < total_pages; i++)
            {
                if (!(page_access_time[i] & PAGE_LOCKED))
                {
                    page_access_time[i] = 0;
                }
            }
        }
        page_access_time[page_index] = current_time;
    }
}

// LRU页面置换算法
void lru_evict_page(void)
{
    uint64_t oldest_time = MAX_TIME;
    uint64_t oldest_page_index = 0;

    for (uint64_t i = 0; i < total_pages; i++)
    {
        if (!(page_access_time[i] & PAGE_LOCKED) && page_access_time[i] < oldest_time)
        {
            oldest_time = page_access_time[i];
            oldest_page_index = i;
        }
    }

    if (oldest_page_index != 0)
    { // 如果找到非锁定的最老页面
        uintptr_t oldest_page_addr = (uintptr_t)(oldest_page_index * 0x1000);
        free_page((void *)oldest_page_addr);
    }
}