// sched.c
#include "sched.h"
#include "video.h"
#include "types.h"
#include "string.h"
#include "memory.h"

static struct task tasks[MAX_PROCESSES];
static struct thread *current_thread = NULL;
static int total_tasks = 0;

ThreadQueue thread_queue;
semaphore_t thread_queue_sem;

// 初始化线程队列
void init_thread_queue(void)
{
    thread_queue.front = 0;
    thread_queue.rear = 0;
    semaphore_init(&thread_queue_sem, 1);
}

// 队列判空
int is_queue_empty(void)
{
    return thread_queue.front == thread_queue.rear;
}

// 入队列操作
int enqueue_thread(struct thread *thread)
{
    if ((thread_queue.rear + 1) % QUEUE_SIZE == thread_queue.front)
    {
        return -1;
    }

    // 找到合适的插入位置
    int insert_pos = thread_queue.front;
    int current_pos = thread_queue.front;

    while (current_pos != thread_queue.rear)
    {
        if (thread_queue.threads[current_pos]->priority < thread->priority)
        {
            break;
        }
        insert_pos = (current_pos + 1) % QUEUE_SIZE;
        current_pos = (current_pos + 1) % QUEUE_SIZE;
    }

    // 移动元素，为新线程腾出位置
    int move_pos = thread_queue.rear;
    while (move_pos != insert_pos)
    {
        int prev_pos = (move_pos - 1 + QUEUE_SIZE) % QUEUE_SIZE;
        thread_queue.threads[move_pos] = thread_queue.threads[prev_pos];
        move_pos = prev_pos;
    }

    thread_queue.threads[insert_pos] = thread;
    thread_queue.rear = (thread_queue.rear + 1) % QUEUE_SIZE;
    return 0;
}

// 出队列操作
struct thread *dequeue_thread(void)
{
    if (thread_queue.front == thread_queue.rear)
    {
        return NULL;
    }
    struct thread *thread = thread_queue.threads[thread_queue.front];
    thread_queue.front = (thread_queue.front + 1) % QUEUE_SIZE;
    return thread;
}

// 老化机制，等待时间过长的线程提升优先级
void update_thread_priority(void)
{
    int current_pos = thread_queue.front;

    while (current_pos != thread_queue.rear)
    {
        struct thread *t = thread_queue.threads[current_pos];
        t->wait_time++;

        // 当等待时间超过阈值时，提升优先级
        if (t->wait_time >= AGING_THRESHOLD)
        {
            int new_priority = min(t->priority + PRIORITY_BOOST, MAX_PRIORITY);
            if (new_priority != t->priority)
            {
                t->priority = new_priority;
                t->wait_time = 0; // 重置等待时间

                // 可能需要重新排序队列
                struct thread *temp = t;
                // 从队列中移除该线程
                int i = current_pos;
                while (i != thread_queue.rear - 1)
                {
                    thread_queue.threads[i] = thread_queue.threads[i + 1];
                    i = (i + 1) % QUEUE_SIZE;
                }
                thread_queue.rear--;
                // 重新插入队列
                enqueue_thread(temp);
            }
        }
        current_pos = (current_pos + 1) % QUEUE_SIZE;
    }
}

// 初始化线程栈
static int init_thread_stack(struct thread_stack *stack)
{
    void *stack_page = alloc_page();
    if (!stack_page)
    {
        return -1;
    }

    stack->stack_page = stack_page;
    stack->stack_top = (uint32_t)stack_page + THREAD_STACK_SIZE;

    return 0;
}

// 释放线程栈
static void free_thread_stack(struct thread_stack *stack)
{
    if (stack->stack_page)
    {
        free_page(stack->stack_page);
        stack->stack_page = NULL;
        stack->stack_top = 0;
    }
}

// 创建线程
int create_thread(int pid, unsigned int eip, int priority, void *arg)
{
    if (pid >= total_tasks || tasks[pid].thread_count >= MAX_THREADS)
    {
        return -1;
    }

    struct task *process = &tasks[pid];
    int thread_id = process->thread_count++;
    struct thread *new_thread = &process->threads[thread_id];

    if (init_thread_stack(&new_thread->stack) != 0)
    {
        process->thread_count--;
        return -1;
    }

    uint32_t *stack_ptr = (uint32_t *)new_thread->stack.stack_top;

    *(--stack_ptr) = (uint32_t)arg;
    *(--stack_ptr) = 0;
    *(--stack_ptr) = eip;
    *(--stack_ptr) = 0;
    *(--stack_ptr) = 0;
    *(--stack_ptr) = 0;
    *(--stack_ptr) = 0;

    new_thread->esp = (uint32_t)stack_ptr;
    new_thread->ebp = 0;
    new_thread->eip = eip;
    new_thread->time_slice = TIME_SLICE;
    new_thread->state = TASK_READY;
    new_thread->thread_id = thread_id;
    new_thread->parent_process = process;
    new_thread->priority = priority;
    new_thread->initial_priority = priority; // 保存初始优先级
    new_thread->wait_time = 0;               // 初始化等待时间

    semaphore_wait(&thread_queue_sem);
    enqueue_thread(new_thread);
    semaphore_signal(&thread_queue_sem);

    return thread_id;
}

// 终止线程
void terminate_thread(int pid, int thread_id)
{
    if (pid >= total_tasks || thread_id >= tasks[pid].thread_count)
    {
        return;
    }

    struct thread *thread = &tasks[pid].threads[thread_id];

    if (thread == current_thread)
    {
        free_thread_stack(&thread->stack);
        thread->state = TASK_BLOCKED;
        schedule();
    }
    else
    {
        free_thread_stack(&thread->stack);
        thread->state = TASK_BLOCKED;
    }
}

// 创建任务进程
int add_task(unsigned int eip, int priority, int remaining_time)
{
    if (total_tasks >= MAX_PROCESSES)
    {
        return -1;
    }

    struct task *new_task = &tasks[total_tasks];
    new_task->pid = total_tasks;
    new_task->priority = priority;
    new_task->remaining_time = remaining_time;
    new_task->thread_count = 0;

    // Allocate a page table for the new task
    void *page_table = os_alloc_page();
    if (!page_table)
    {
        return -1;
    }

    new_task->page_table_phys_addr = (uint32_t)(uintptr_t)page_table;

    if (create_thread(total_tasks, eip, priority, NULL) < 0)
    {
        free_page(page_table);
        return -1;
    }

    int pid = total_tasks;
    total_tasks++;
    return pid;
}

// 线程调度（基于优先级队列的时间片轮转算法，由硬件定时器中断支持）
void schedule(void)
{
    semaphore_wait(&thread_queue_sem);

    // 更新所有等待线程的优先级
    update_thread_priority();

    if (!current_thread || current_thread->time_slice <= 0 ||
        current_thread->state != TASK_RUNNING)
    {

        struct thread *next_thread = dequeue_thread();
        if (!next_thread)
        {
            semaphore_signal(&thread_queue_sem);
            return;
        }

        if (current_thread)
        {
            current_thread->esp = get_esp();
            current_thread->ebp = get_ebp();
            current_thread->eip = get_eip();

            if (current_thread->state == TASK_RUNNING)
            {
                current_thread->state = TASK_READY;
                current_thread->wait_time = 0; // 重置等待时间

                // 如果优先级被提升过，逐步降回初始优先级
                if (current_thread->priority > current_thread->initial_priority)
                {
                    current_thread->priority--;
                }

                enqueue_thread(current_thread);
            }
        }

        current_thread = next_thread;
        current_thread->state = TASK_RUNNING;
        current_thread->time_slice = TIME_SLICE;
        current_thread->wait_time = 0; // 重置等待时间

        switch_context(current_thread->esp,
                       current_thread->ebp,
                       current_thread->eip);
    }

    semaphore_signal(&thread_queue_sem);
}

// 消耗线程时间片的一个tick，时间片消耗完则触发schedule()进行调度
void timer_tick(void)
{
    if (current_thread)
    {
        current_thread->time_slice--;
        if (current_thread->time_slice <= 0)
        {
            schedule();
        }
    }
}

// 让当前线程放弃处理器
void yield_thread(void)
{
    if (current_thread)
    {
        current_thread->time_slice = 0;
        schedule();
    }
}

// 阻塞当前线程
void block_current_thread(void)
{
    if (current_thread)
    {
        current_thread->state = TASK_BLOCKED;
        schedule();
    }
}

// 激活线程
void unblock_thread(struct thread *thread)
{
    if (thread)
    {
        thread->state = TASK_READY;
        semaphore_wait(&thread_queue_sem);
        enqueue_thread(thread);
        semaphore_signal(&thread_queue_sem);
    }
}

// 初始化信号量
void semaphore_init(semaphore_t *sem, int value)
{
    sem->value = value;
    sem->waiters = &thread_queue; // 使用全局任务队列作为信号量的等待队列
}

// 等待信号量（P操作）
void semaphore_wait(semaphore_t *sem)
{
    // 先检查信号量的值
    if (sem->value > 0)
    {
        sem->value--; // 信号量值大于0，直接减少
    }
    else
    {
        sem->highest_priority_waiter = max(sem->highest_priority_waiter, current_thread->priority);
        // 信号量值为0，当前线程需要被阻塞
        if (is_queue_empty())
        {
            // 如果当前线程是唯一线程，设置为TASK_SLEEPING
            current_thread->state = TASK_SLEEPING; // 将当前线程状态设置为睡眠
            schedule();                            // 调度器将选择其他线程运行
        }
        else
        {
            // 将当前线程加入等待队列
            enqueue_thread(current_thread);
            block_current_thread(); // 阻塞当前线程
        }
    }
}

// 释放信号量（V操作）
void semaphore_signal(semaphore_t *sem)
{
    sem->value++; // 增加信号量值

    if (sem->value <= 0)
    {
        // 有线程在等待这个信号量
        struct thread *thread_to_wake = dequeue_thread(); // 从等待队列中取出一个线程
        if (thread_to_wake)
        {
            thread_to_wake->state = TASK_READY; // 改变线程状态为就绪
            if (thread_to_wake->priority < sem->highest_priority_waiter)
            {
                thread_to_wake->priority = sem->highest_priority_waiter;
            }
            // 将线程加入到就绪队列中，以便调度器可以调度它
            enqueue_thread(thread_to_wake);
        }
    }
}
